'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.observer = undefined;

var _slicedToArray2 = require('babel-runtime/helpers/slicedToArray');

var _slicedToArray3 = _interopRequireDefault(_slicedToArray2);

var _extends2 = require('babel-runtime/helpers/extends');

var _extends3 = _interopRequireDefault(_extends2);

var _react = require('react');

var React = _interopRequireWildcard(_react);

var _observerUtil = require('@nx-js/observer-util');

var _batching = require('../../batching');

var _batching2 = _interopRequireDefault(_batching);

var _destroyer = require('./destroyer');

var _destroyer2 = _interopRequireDefault(_destroyer);

var _promiseBatcher = require('./promiseBatcher');

var _promiseBatcher2 = _interopRequireDefault(_promiseBatcher);

var _model = require('@startupjs/model');

var _model2 = _interopRequireDefault(_model);

var _meta = require('./meta');

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var DEFAULT_OPTIONS = {
  forwardRef: false,
  suspenseProps: {
    fallback: React.createElement(NullComponent, null, null)
  }

  // TODO: Fix passing options argument in react-native Fast Refresh patch.
  //       It has to properly put the closing bracket.
}; // ref: https://github.com/mobxjs/mobx-react-lite/blob/master/src/observer.ts
function observer(Component, options) {
  var _options = (0, _extends3.default)({}, DEFAULT_OPTIONS, options);
  return wrapObserverMeta(makeObserver(Component, _options), _options);
}

observer.__wrapObserverMeta = wrapObserverMeta;
observer.__makeObserver = makeObserver;

exports.observer = observer;


function pipeComponentMeta(SourceComponent, TargetComponent) {
  var suffix = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : '';
  var defaultName = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : 'StartupjsWrapper';

  var displayName = SourceComponent.displayName || SourceComponent.name;
  if (!TargetComponent.displayName) {
    TargetComponent.displayName = displayName ? displayName + suffix : defaultName;
  }
  if (!TargetComponent.propTypes && SourceComponent.propTypes) {
    TargetComponent.propTypes = SourceComponent.propTypes;
  }
  if (!TargetComponent.defaultProps && SourceComponent.defaultProps) {
    TargetComponent.defaultProps = SourceComponent.defaultProps;
  }
  return TargetComponent;
}

function makeObserver(baseComponent) {
  var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

  var _DEFAULT_OPTIONS$opti = (0, _extends3.default)({}, DEFAULT_OPTIONS, options),
      forwardRef = _DEFAULT_OPTIONS$opti.forwardRef;
  // MAGIC. This fixes hot-reloading. TODO: figure out WHY it fixes it


  var random = Math.random();

  // memo; we are not intested in deep updates
  // in props; we assume that if deep objects are changed,
  // this is in observables, which would have been tracked anyway
  var WrappedComponent = function WrappedComponent() {
    // forceUpdate 2.0
    var forceUpdate = useForceUpdate();

    // wrap the baseComponent into an observe decorator once.
    // This way it will track any observable changes and will trigger rerender
    var observedComponent = React.useMemo(function () {
      var blockUpdate = { value: false };
      var update = function update() {
        // TODO: Decide whether the check for unmount is needed here
        // Force update unless update is blocked. It's important to block
        // updates caused by rendering
        // (when the sync rendening is in progress)
        if (!blockUpdate.value) forceUpdate();
      };
      var batchedUpdate = function batchedUpdate() {
        return _batching2.default.add(update);
      };
      return (0, _observerUtil.observe)(wrapBaseComponent(baseComponent, blockUpdate), {
        scheduler: batchedUpdate,
        lazy: true
      });
    }, [random]);

    // clean up observer on unmount
    useUnmount(function () {
      return (0, _observerUtil.unobserve)(observedComponent);
    });

    return observedComponent.apply(undefined, arguments);
  };

  pipeComponentMeta(baseComponent, WrappedComponent);
  return forwardRef ? React.forwardRef(WrappedComponent) : WrappedComponent;
}

function wrapObserverMeta(Component) {
  var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

  var _DEFAULT_OPTIONS$opti2 = (0, _extends3.default)({}, DEFAULT_OPTIONS, options),
      forwardRef = _DEFAULT_OPTIONS$opti2.forwardRef,
      suspenseProps = _DEFAULT_OPTIONS$opti2.suspenseProps;

  if (!(suspenseProps && suspenseProps.fallback)) {
    throw Error('[observer()] You must pass at least a fallback parameter to suspenseProps');
  }

  function ObserverWrapper(props, ref) {
    var componentMeta = React.useMemo(function () {
      return {
        componentId: _model2.default.id(),
        createdAt: Date.now()
      };
    }, []);
    return React.createElement(_meta.ComponentMetaContext.Provider, { value: componentMeta }, React.createElement(React.Suspense, suspenseProps, forwardRef ? React.createElement(Component, (0, _extends3.default)({}, props, { ref: ref })) : React.createElement(Component, props)));
  }

  var memoComponent = void 0;
  pipeComponentMeta(Component, ObserverWrapper, 'Observer', 'StartupjsWrapperObserver');

  if (forwardRef) {
    memoComponent = React.memo(React.forwardRef(ObserverWrapper));
  } else {
    memoComponent = React.memo(ObserverWrapper);
  }

  pipeComponentMeta(Component, memoComponent);
  return memoComponent;
}

function wrapBaseComponent(baseComponent, blockUpdate) {
  return function () {
    blockUpdate.value = true;
    var res = void 0;
    try {
      _destroyer2.default.reset();
      _promiseBatcher2.default.reset();
      res = baseComponent.apply(undefined, arguments);
    } catch (err) {
      if (!err.then) throw err;
      // If the Promise was thrown, we catch it before Suspense does.
      // And we run destructors for each hook previous to the one
      // which did throw this Promise.
      // We have to manually do it since the unmount logic is not working
      // for components which were terminated by Suspense as a result of
      // a promise being thrown.
      var destroy = _destroyer2.default.getDestructor();
      throw err.then(destroy);
    }
    blockUpdate.value = false;
    if (_promiseBatcher2.default.isActive()) {
      throw Error('[react-sharedb] useBatch* hooks were used without a closing useBatch() call.');
    }
    return res;
  };
}

function useForceUpdate() {
  var _React$useState = React.useState(),
      _React$useState2 = (0, _slicedToArray3.default)(_React$useState, 2),
      setTick = _React$useState2[1];

  return function () {
    setTick(Math.random());
  };
}

// TODO: Might change to just `useEffect` in future. Don't know which one fits here better yet.
function useUnmount(fn) {
  React.useLayoutEffect(function () {
    return fn;
  }, []);
}

function NullComponent() {
  return null;
}